use std::rc::Rc;

use crate::func::*;

pub type NodeRc<'a, BT> = Rc<<BT as BinaryTrees<'a>>::Node>;
pub type ReferenceRc<'a, BT> = Rc<<BT as BinaryTrees<'a>>::Reference>;
pub type TreeRc<'a, BT> = Rc<<BT as BinaryTrees<'a>>::Tree>;
pub type KeyRc<'a, BT> = Rc<<BT as BinaryTrees<'a>>::Key>;

pub type Split<'a, BT> = (TreeRc<'a, BT>, TreeRc<'a, BT>, KeyRc<'a, BT>);
pub type KeySplit<'a, BT> = (TreeRc<'a, BT>, TreeRc<'a, BT>);

pub type Wrapped<'a, BT, A> = Wrap<'a, A, <BT as BinaryTrees<'a>>::T>;

pub trait BinaryTrees<'a>: 'a {
    type Node: 'a;
    type Reference: 'a;
    type Tree: 'a;
    type Key: 'a;

    type T: Monad<'a>;

    fn split(node: Self::Node) -> Wrapped<'a, Self, Split<'a, Self>>;
    fn to_tree(node: Self::Node) -> TreeRc<'a, Self>;
    fn to_tree_construct(&self, node: Self::Node) -> TreeRc<'a, Self> {
        Self::to_tree(node)
    }
    fn resolve(reference: Self::Reference) -> Wrapped<'a, Self, NodeRc<'a, Self>>;
    fn equal(rhs: Self::Reference, lhs: Self::Reference) -> bool;
    fn refer(tree: Self::Tree) -> Option<ReferenceRc<'a, Self>>;
    fn join_key(
        tl: Self::Tree,
        key: KeyRc<'a, Self>,
        tr: Self::Tree,
    ) -> Wrapped<'a, Self, TreeRc<'a, Self>>;
    fn join(tl: Self::Tree, tr: Self::Tree) -> Wrapped<'a, Self, TreeRc<'a, Self>>;
    fn split_key(tree: Self::Tree, key: KeyRc<'a, Self>) -> Wrapped<'a, Self, KeySplit<'a, Self>>;
}
