use crate::flow::traversible::*;

pub fn n_contains<'a, T: MonadFail<'a, ()>, A: 'a, D: 'a + PartialEq>(
    comparator: &'a dyn Comparator<A>,
    n_set: Rc<dyn TraversibleBinaryNode<'a, T, A, D>>,
    key: Rc<A>,
) -> T::F<()> {
    let (t_setl, t_setr, k_set) = n_set.split();
    match comparator.pick_smaller(key.as_ref(), k_set.as_ref()) {
        Comparison::L => t_contains(comparator, t_setl, key),
        Comparison::E => T::pure(()),
        Comparison::R => t_contains(comparator, t_setr, key),
    }
}

pub fn r_contains<'a, T: MonadFail<'a, ()>, A: 'a, D: 'a + PartialEq>(
    comparator: &'a dyn Comparator<A>,
    r_set: Rc<dyn TraversibleBinaryReference<'a, T, A, D>>,
    key: Rc<A>,
) -> T::F<()> {
    T::bind(r_set.resolve(), |n_set| n_contains(comparator, n_set, key))
}

pub fn t_contains<'a, T: MonadFail<'a, ()>, A: 'a, D: 'a + PartialEq>(
    comparator: &'a dyn Comparator<A>,
    t_set: Rc<dyn TraversibleBinaryTree<'a, T, A, D>>,
    key: Rc<A>,
) -> T::F<()> {
    match t_set.refer() {
        Some(r_set) => r_contains(comparator, r_set, key),
        None => T::fail(()),
    }
}
